""" 

pyFaspMgmt.py -- A python implementation of the Aspera FASP Manager Interface.
  (c) IBM 2015-16

""" 

import conf
import log
import signal
import time
import socket 
import os
import itertools
import tempfile
import subprocess
import pprint
import glob
import Queue
import threading 

transferId=itertools.count()
activeTransfers={}
mgmt_lock = threading.Condition() 
mgmt_port = 0

class FaspSend:
  """
  A python FASP Management Interface.
  """

  def __init__(self, 
    user, dest_host, 
    src_files=[], dest_files=[], args=[], dest_dir=".",
    password="", key_file="", token="", filepass="", proxy_pass="",
    persistent=False, receiver=False, ascpPath=None, ascpRoot=None
  ):
    """
      user:        username to use when connecting to remote server
      dest_host:   destination server to connect to 
      src_files:   files to transfer (needed for non persistent sessions)
      dest_files:  filenames to use at destination
      dest_dir:    directory to store files 
      args:        ASCP command line arguments (see ascp -h)
      password:    SSH Password 
      key_file:    Private SSH key used to connect to destination server 
      token:       ASCP token used to connect to destination server
      filepass:    Key used for encryption at rest
      proxy_pass:  Password for proxy (specify proxy in args)
      persistent:  Start a persistent session
      receiver:    Start as receiver
      ascpPath:    ASCP Executable path
      ascpRoot:    Path to ASCP directory
    """

    def getPyFaspId():
      return "pyFM_%s.%x.%x" % ( conf.cfg.get("all", "sysId"),
                                 int(conf.cfg.get("all", "sessionNo")),
                                 transferId.next() )

    def build_filelist( self, src, dst, args ):
      with tempfile.NamedTemporaryFile( delete=False ) as tmp:
        fi_list = []
        if src and dst:
          args.append("--file-pair-list") 
          for i in zip(src, dst):
            fi_list = fi_list + list(i)
        else:
          args.append("--file-list") 
          fi_list = src
        tmp.write("\n".join(fi_list))
        tmp.write("\n")
        self.file_list=tmp.name
        args.append(self.file_list)


    def ascp_thread( args, env, queue ):
      try: 
        env.update(os.environ)  
        output = subprocess.check_output( args, env=env, stderr=subprocess.STDOUT )
        queue.put(None)
        log.info("ASCP exited normally (%s)" % output.strip())
      except subprocess.CalledProcessError, e:
        msg = "ASCP exited with errorcode=%d. " % e.returncode
        if e.output.strip():
          msg = "%s; Error was \"%s\"" % ( msg, e.output.strip() )
        log.info(msg)
        queue.put(msg)
      except Exception, e:
        msg = "Error calling ASCP (%s); %s" % (args[0], e)
        log.err(msg)
        queue.put(msg)

    self.pct = 0
    self.bytesTransferred = 0
    self.keepalive = False
    self.ascp_isrunning = False
    self.files = src_files
    self.dest_host = dest_host
    env = {}
    self.stat = "Not Started"
    self.totalBytes = 0
    self.keepalive = False
    self.ident = "Not Assigned"
    self.stat = "Not Started"
    self.dirty = True

    self.filesComplete = {}
    self.errs = []
    self.totalBytes = 0


    if not src_files and not persistent:
      raise ValueError("No files specified")

    if src_files and persistent:
      raise ValueError("Can not specify src_files with persistent")

    if ascpRoot: 
      log.dbg("pyfaspmgmt: set ascpRoot to %s" % ascpRoot)

    if ascpPath:
      args = [ ascpPath, ] + args
    else:
      args = [ conf.cfg.get("all", "ascpPath") ] + args

    if (key_file): 
      args.append("-i")
      args.append(key_file)

    if (password):
      env['ASPERA_SCP_PASS']=password

    if (token):
      env['ASPERA_SCP_TOKEN']=token

    if (filepass):
      env['ASPERA_SCP_FILEPASS']=filepass

    if (proxy_pass):
      env['ASPERA_PROXY_PASS']=proxy_pass

    if (persistent):
      args.append('--keepalive')
      self.keepalive = True

    global mgmt_port, mgmt_lock
    if mgmt_port < 1:
      log.dbg("Acquire management lock")
      mgmt_lock.acquire() 
      if not mgmt_port: 
        log.dbg("start_fasp_mgmt")
        mgmt_port = -1
        t = threading.Thread( target=start_fasp_mgmt )
        t.daemon = True
        t.start()
      log.dbg("Wait on management lock")
      mgmt_lock.wait()
      log.dbg("Release management lock")
      mgmt_lock.release()
      
    args.append("-q")
    args.append("-M")
    args.append(str(mgmt_port))
    args.append("-u")
    self.ident= getPyFaspId()
    args.append(self.ident)
    # for 2.x;
    #args.append("%s@%s:%s" % (user, dest_host, dest_dir) )  

    # for 3.x;
    args.append('--mode')
    if not receiver:
      args.append('SEND')
    else:
      args.append('RECV')

    args.append('--host')
    args.append(dest_host)
    args.append('--user')
    args.append(user)
    if not persistent: 
      build_filelist( self, src_files, dest_files, args )
    args.append ( dest_dir )
    log.info("(%s) File list = [%s]" % (self.ident, ",".join(src_files)) )

    log.info("(%s) Calling ascp with %s" % (self.ident, " ".join(args)) )

    self.ascp = Queue.Queue()
    self.ascp_isrunning=True
    thread = threading.Thread( target=ascp_thread, args=(args, env, self.ascp) )
    thread.daemon = True
    thread.start()

  def consume_msg(self):
    """ Internal-- Parses a single message from FaspMgmt Connection """
    while 1: 
      gotMessage = False
      while gotMessage is not True:
        try: 
          msg = activeTransfers[self.ident][0].get_nowait()
        except Exception, e:
          #log.dbg("consume_msg: %s" % e)
          return # No messages to consume
        msgTypes = ( "INIT", "SESSION", "STATS", "STOP", "DONE", "FILEERROR", "ERROR", "NOTIFICATION" )
        log.dbg("Got message of type %s" % msg["Type"] )

        if (msg["Type"] not in msgTypes):
          log.dbg("Ignoring message (type='%s')" % msg["Type"])
        else:
          gotMessage = True

      try: 
        self.bytesTransferred = int(msg["FileBytes"]) 
      except:
        pass
     
      if msg["Type"] in ( "NOTIFICATION" ):
        self.notif = msg
        try:
          if not self.keepalive: 
            self.totalBytes += int(msg["PreTransferBytes"])
        except:
          pass

      if msg["Type"] in ( "INIT", "SESSION", "DONE", "ERROR" ):
        self.stat = msg["Type"].strip()

      if msg["Type"] in ("DONE", "ERROR"):
        log.info("(%s) Session Complete" % self.ident)
        self.end = msg

      if msg["Type"] in ( "SESSION" ):
        self.params = msg

      if msg["Type"] in ( "ERROR", "FILEERROR" ):
        if not msg.has_key("Description"):
          msg["Description"] = "Error encountered but no error message provided"
        if msg["Type"] == "FILEERROR":
          try: 
            self.filesComplete[msg["FaspFileID"]] = {'name': msg["File"], 'status': 'Fail', 'error': msg["Description"]}
          except:
            self.filesComplete[msg['File']] = {'name': msg["File"], 'status': 'Fail', 'error': msg["Description"]}
          self.errs.append( "File '%s' %s" % (msg["File"], msg["Description"]) )
        else:
          log.info("(%s) Session Failure: %s" % (self.ident, msg["Description"]) )
          self.errs.append( "Session failed: %s" % (msg["Description"]) )
        
      if msg["Type"] in ( "STOP" ):
        try: 
          self.filesComplete[msg["FaspFileID"]] = {'name': msg["File"], 'status': 'Success'}
        except:
          self.filesComplete[msg["File"]] = {'name': msg["File"], 'status': 'Success'}
        try: 
          self.transfersAttempted = msg["TransfersAttempted"]
          self.transfersPassed = msg["TransfersPassed"] 
        except:
          pass


      if msg["Type"] in ( "STATS" ):
        self.stats = msg


  def status(self):
    "Returns the current session state, ( INIT, SESSION, DONE, or ERROR )."
    log.dbg("Doing self.status (%s)" % self.stat )
    self.consume_msg()

    try: 
      res = self.ascp.get_nowait()
      self.ascp_isrunning = False

      if res:
        self.stat = "ERROR"
        self.errs.append(res)

      if res == None:
        # Make sure we have consumed all messages.
        self.consume_msg()
        time.sleep(0.2)
        self.consume_msg()

        if self.stat in ("ERROR", "DONE"):
          return self.stat

        try: 
          if self.transfersAttempted == self.transfersPassed:
            log.dbg("Setting state to DONE based on file success")
            "This is a work around for ASCP4 which does not always print DONE"
            self.stat = "DONE"
          return self.stat
        except:
          log.dbg("Could not get completed file count")
        log.dbg("Setting state to error based on inconclusive xfer results")
        self.errs.append("DONE/ERROR status message not recieved")
        self.stat = "ERROR"
      else:
        log.dbg("Unexpected Queue Value.")
 
    except:
      pass

    return self.stat

  def getParam(self, param):
    "Get a identifier from SESSION notification."
    try:
      self.params["param"]
    except:
      return ""

  def addFile(self, src, dest, depth=0):
    self.files.append(src)
    "If persistent sessions is enabled transfer src and name it dest"
    if not self.keepalive:
      raise ValueError("Start a persistent session to enable addFile support")

    msg = \
"""FASPMGR 2
Type: START
Source: %s
Destination: %s

""" % (src, dest)
    try:
      activeTransfers[self.ident][1].put(msg)
    except Exception, e:
      if (depth > 10):
        raise e
      time.sleep(0.35)
      self.addFile(src,dest,depth+1)


  def closePersistentSession( self ):
    "Close persistent session. Use getStatus to check if session has completed."
    if not self.keepalive:
      raise ValueError("Session is not a persistent session")
    msg = \
"""FASPMGR 2
Type: DONE
Operation: Linger

"""
    activeTransfers[self.ident][1].put(msg)
 
  def pctComplete(self):
    "If precompute is enabled return the percent complete from 0 to 100"
    if self.status() == "DONE":
      return 100

    try:
      pct = (int(self.bytesTransferred) * 100) / self.totalBytes  
      if pct > 100:
        pct = 99
      return pct
    except Exception, e:
      return 0

  def isRunning(self):
    if self.status() in ("DONE", "ERROR"):
      return False
    return True

  def cancel(self):
    if not self.isRunning():
      return False
      
    try:
      # Request that ASCP kill itself.
      msg = \
"""FASPMGR 2
Type: CANCEL

"""
      activeTransfers[self.ident][1].put(msg)
      depth = 0
      while (depth < 10):
        if not self.isRunning(): 
          return True
        time.sleep(0.5)
        depth += 1
      return False
    except Exception, e:
      log.dbg("Got an error trying to kill ASCP: %s" % e)
      pass


  def cleanup(self):
    log.dbg("Called cleanup on %s" % self.ident)
    if self.dirty:
      self.cancel()
      try:
        self.dirty = False
        try:
          if self.file_list:
            pass
            #os.remove(self.file_list)
        except:
          pass
        self.ascp.kill()
         
      except:
        pass
     

  def close(self):
    "Terminate all resources associated with transfer including active transfer."
    self.cleanup()

  def __del__(self):
    self.cleanup()

  def __exit__(self, type, value, traceback):
    self.cleanup()




def start_fasp_mgmt( port=33500, host="localhost", ascpRoot=""):
  """ Internal function-- called when first FaspSend object is initialized """
  def mgmt_connection( sock, path ):
    def centralConnect(): 
      port_paths = glob.glob("%s/var/run/aspera/asperacentral.port" % path) + \
                   glob.glob("%s/var/run/aspera/*.optport" % path)


      central = [] 

      for port_fn in port_paths: 
        try: 
          portnum = 0
          log.dbg("Reading port %s" % port_fn)
          with open(port_fn) as cPort:
             portnum = int(cPort.read())
          c= socket.socket()
          c.connect( ('127.0.0.1', int(portnum) ) )
          central.append(c)
          log.dbg("Success connecting to central")
        except Exception, e:
          log.err("Connecting to port %s; %s" % ( port_fn, str(e)))

      return central

    sock.settimeout(.5)
    notifications=[]
    uid = ""
    sessionActive = True
    data = ""


    closeMgmt = False
    central = []

    while 1:
      msg = {}
      try: 
        if (notifications): 
          if 'FASP' not in notifications[0]:
            log.err("BAD notif: %s" % notifications[0])
            notifications = notifications[1:]
            continue
          for i in range( 1, len(notifications) ):
            line = notifications[i]
            if line: 
              try: 
                k,v = line.split(": ", 1)
              except:
                log.dbg("Bad line: %s" % line)
                continue
              msg[k]=v
            else:
              fm2 = "%s\n" %  "\n".join(notifications) 
              #print "****************\n%s************" % fm2
              try: 
                central = centralConnect()
              except Exception, e:
                pass

              for i in central: 
                try: 
                  i.sendall(fm2)
                except Exception, e: 
                  pass
              #notifications=notifications[i+1:]
              log.dbg("[%s] message from ASCP %s" % (msg['UserStr'], pprint.pformat(msg)) )
              try:
                if (msg): 
                  activeTransfers[msg['UserStr']][0].put(msg)
              except:
                try:
                  # Create two queues, one for messages received from fasp mgmt,
                  # the other for messages to send to fasp mgmt.
                  activeTransfers[msg['UserStr']] = ( Queue.Queue(), Queue.Queue() )
                  activeTransfers[msg['UserStr']][0].put(msg)
                  uid = msg['UserStr']
                except Exception, e:
                  log.info("No UserId specified for transfer, closing FaspMgmt. %s" % str(e) )
                  sessionActive = False
              msg = {}
              continue # Parsing notifications until all notifications consumed.
      except Exception, e:
        print "NOTIF ERR", e
              
              

      if uid:
        msg = None
        try: 
          msg = activeTransfers[uid][1].get(True, 0.0001)
        except Queue.Empty:
          pass
        except Exception, e:
          print "Queue.Exception", e
        try: 
          if msg:
            log.dbg("Sending msg to ascp '%s'" % msg)
            sock.sendall(msg)
            continue
        except: 
          pass
            
      if closeMgmt:
        log.dbg("Closing mgmt connection; socket closed")     
        return

      gotData = False
      gotTimeout = False
      notifications = []

      h = ""
      try: 
        if "\n\n" in data:
          messages = data.split("\n\n", 1);
          data = messages [-1]
          for i in messages[:-1]:
            notifications = i.splitlines()
            notifications.append("")
          continue
        h = sock.recv(1024)
        if h:
          gotData = True
        data += h

      except socket.timeout, e:
        gotTimeout = True
        pass
      except Exception, e:
        log.dbg("Closing mgmt connection; %s" % e)
        sessionActive = False
        return

      if not gotData and not gotTimeout:
        closeMgmt = True

  class Listener:
    t = []

    def __init__(self, host, port, function, ascpRoot):
      self.s = socket.socket()
      #self.s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
      self.s.bind( (host, port) )
      self.s.listen(500)
      self.fun = function

    def accept(self):
      try:
        newsock = self.s.accept()
        log.dbg("Received connection from %s" % newsock[1][0])
        thread = threading.Thread (target=self.fun, args=(newsock[0],ascpRoot) )
        thread.daemon = True
        thread.start()
      except Exception, e:
        log.dbg( "On Receive: %s" % e )

  listener = ""

  try:
    port = conf.cfg.get("all", "mgmtPort")
    host = conf.cfg.get("all", "mgmtHost")
  except:
    pass

  tries = 0
  while tries < 10:
    try:
      listener = Listener( host, port+tries, mgmt_connection, ascpRoot )
      break
    except:
      tries+=1

  if tries > 9:
    log.err("Couldn't listen on management port")
    raise ValueError("Couldn't listen on management port")


  mgmt_lock.acquire()

  global mgmt_port
  mgmt_port = port+tries
  log.dbg("Set management port to %d" % mgmt_port)

  mgmt_lock.notify()
  mgmt_lock.release()

  log.log("Starting pyFaspMgmt git %s built %s on port %d" % (conf.VERSION, conf.DATE, port), "info")

  while 1:
    listener.accept( )





